"""A flexible caching decorator.

This module provides a cache decorator `cache` that you can use to
cache results of your functions or methods.
"""

from functools import wraps

import time


class CleanupDict(dict):
    """A dict that automatically cleans up items that haven't been
    accessed in a given timespan on *set*.
    """

    cleanup_period = 60 * 60 * 24 * 3  # 3 days

    def __init__(self, cleanup_period=None):
        super().__init__()
        self._last_access = {}
        if cleanup_period is not None:
            self.cleanup_period = cleanup_period

    def __getitem__(self, key):
        value = super().__getitem__(key)
        self._last_access[key] = time.time()
        return value

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self._last_access[key] = time.time()
        self._cleanup()

    def _cleanup(self):
        now = time.time()
        okay = now - self.cleanup_period
        for key, timestamp in list(self._last_access.items()):
            if timestamp < okay:
                del self._last_access[key]
                super().__delitem__(key)


ATTR = "_v_memoize_cache"
CONTAINER_FACTORY = CleanupDict
_marker = object()


class DontCache(Exception):
    pass


def store_on_self(method, obj, *args, **kwargs):
    return obj.__dict__.setdefault(ATTR, CONTAINER_FACTORY())


def store_on_context(method, obj, *args, **kwargs):
    return obj.context.__dict__.setdefault(ATTR, CONTAINER_FACTORY())


def cache(get_key, get_cache=store_on_self):
    def decorator(fun):
        @wraps(fun)
        def replacement(*args, **kwargs):
            try:
                key = get_key(fun, *args, **kwargs)
            except DontCache:
                return fun(*args, **kwargs)
            key = f"{fun.__module__}.{fun.__name__}:{key}"
            cache = get_cache(fun, *args, **kwargs)
            cached_value = cache.get(key, _marker)
            if cached_value is _marker:
                cached_value = cache[key] = fun(*args, **kwargs)
            return cached_value

        return replacement

    return decorator



"""Memo decorators for globals - memoized values survive for as long as the
process lives.

Stores values in a module-level variable.

Pay attention that is module is not thread-safe, so use it with care.
"""

from plone.memoize import volatile


_memos = {}


def memoize(fun):
    def get_key(fun, *args, **kwargs):
        return (
            args,
            frozenset(kwargs.items()),
        )

    def get_cache(fun, *args, **kwargs):
        return _memos

    return volatile.cache(get_key, get_cache)(fun)


__all__ = (memoize,)



"""A cache decorator that uses RAMCache by default.
"""

from hashlib import sha1
from plone.memoize import volatile
from plone.memoize.interfaces import ICacheChooser
from zope import component
from zope import interface
from zope.ramcache import ram
from zope.ramcache.interfaces.ram import IRAMCache

import pickle


global_cache = ram.RAMCache()
global_cache.update(maxAge=86400)

DontCache = volatile.DontCache
MARKER = object()


class AbstractDict:
    def get(self, key, default=None):
        try:
            return self.__getitem__(key)
        except KeyError:
            return default


class MemcacheAdapter(AbstractDict):
    def __init__(self, client, globalkey=""):
        self.client = client
        self.globalkey = globalkey and "%s:" % globalkey

    def _make_key(self, source):
        if issubclass(type(source), str):
            source = source.encode("utf-8")
        return sha1(source).hexdigest()

    def __getitem__(self, key):
        cached_value = self.client.get(self.globalkey + self._make_key(key))
        if cached_value is None:
            raise KeyError(key)
        else:
            return pickle.loads(cached_value)

    def __setitem__(self, key, value):
        cached_value = pickle.dumps(value)
        self.client.set(self.globalkey + self._make_key(key), cached_value)


class RAMCacheAdapter(AbstractDict):
    def __init__(self, ramcache, globalkey=""):
        self.ramcache = ramcache
        self.globalkey = globalkey

    def _make_key(self, source):
        if issubclass(type(source), str):
            source = source.encode("utf-8")
        return sha1(source).digest()

    def __getitem__(self, key):
        value = self.ramcache.query(
            self.globalkey, dict(key=self._make_key(key)), MARKER
        )
        if value is MARKER:
            raise KeyError(key)
        else:
            return value

    def __setitem__(self, key, value):
        self.ramcache.set(value, self.globalkey, dict(key=self._make_key(key)))


def choose_cache(fun_name):
    return RAMCacheAdapter(component.queryUtility(IRAMCache), globalkey=fun_name)


interface.directlyProvides(choose_cache, ICacheChooser)


def store_in_cache(fun, *args, **kwargs):
    key = f"{fun.__module__}.{fun.__name__}"
    cache_chooser = component.queryUtility(ICacheChooser)
    if cache_chooser is not None:
        return cache_chooser(key)
    else:
        return RAMCacheAdapter(global_cache, globalkey=key)


def cache(get_key):
    return volatile.cache(get_key, get_cache=store_in_cache)



"""Memo decorators for instances.

Stores values in an attribute on the instance. See instance.rst.

This package current subsumes memojito.
"""
from functools import wraps


_marker = object()


class Memojito:
    propname = "_memojito_"

    def clear(self, inst):
        if hasattr(inst, self.propname):
            delattr(inst, self.propname)

    def clearbefore(self, func):
        def clear(*args, **kwargs):
            inst = args[0]
            self.clear(inst)
            return func(*args, **kwargs)

        return clear

    def clearafter(self, func):
        def clear(*args, **kwargs):
            inst = args[0]
            val = func(*args, **kwargs)
            self.clear(inst)
            return val

        return clear

    def memoize(self, func):
        @wraps(func)
        def memogetter(*args, **kwargs):
            inst = args[0]
            cache = getattr(inst, self.propname, _marker)
            if cache is _marker:
                setattr(inst, self.propname, dict())
                cache = getattr(inst, self.propname)

            # XXX this could be potentially big, a custom key should
            # be used if the arguments are expected to be big

            key = (func.__name__, args, frozenset(kwargs.items()))
            val = cache.get(key, _marker)
            if val is _marker:
                val = func(*args, **kwargs)
                cache[key] = val
                setattr(inst, self.propname, cache)
            return val

        return memogetter


_m = Memojito()
memoize = _m.memoize
clearbefore = _m.clearbefore
clearafter = _m.clearafter


def memoizedproperty(func):
    return property(_m.memoize(func))


__all__ = (memoize, memoizedproperty, clearbefore, clearafter)



"""XHTML Compressor
"""

from plone.memoize.interfaces import IXHTMLCompressor
from zope.component import queryUtility


def xhtml_compress(string):
    util = queryUtility(IXHTMLCompressor)
    if util is not None:
        return util.compress(string)
    return string



"""Memoize decorator for methods.

Stores values in an annotation of the request.
"""
from functools import wraps
from inspect import getfullargspec
from plone.memoize import volatile
from zope.annotation.interfaces import IAnnotations


_marker = object()


class RequestMemo:
    key = "plone.memoize_request"

    def __init__(self, arg=0):
        self.arg = arg

    def __call__(self, func):
        @wraps(func)
        def memogetter(*args, **kwargs):
            request = None
            if isinstance(self.arg, int):
                request = args[self.arg]
            else:
                request = kwargs[self.arg]

            annotations = IAnnotations(request)
            cache = annotations.get(self.key, _marker)

            if cache is _marker:
                cache = annotations[self.key] = dict()

            key = (
                func.__module__,
                func.__name__,
                args,
                frozenset(list(kwargs.items())),
            )
            value = cache.get(key, _marker)
            if value is _marker:
                value = cache[key] = func(*args, **kwargs)
            return value

        return memogetter


def store_in_annotation_of(expr):
    def _store_in_annotation(fun, *args, **kwargs):
        # Use expr to find out the name of the request variable
        vars = {}
        spec = getfullargspec(fun)
        num_args = len(args)
        expected_num_args = num_args

        # Explicitly check for the correct number of arguments and
        # raise the appropriate TypeError if needed. This is done
        # to avoid the real problem being masked by an IndexError
        # later in this method.
        if spec[3] is not None:
            expected_num_args = len(spec[0]) - len(spec[3])
        if num_args != expected_num_args:
            raise TypeError(
                "%s() takes exactly %s arguments (%s given)"
                % (fun.__name__, expected_num_args, num_args)
            )

        for index, name in enumerate(spec[0]):
            if index < num_args:
                vars[name] = args[index]
            else:
                vars[name] = kwargs.get(name, spec[3][index - num_args])
        request = eval(expr, {}, vars)
        return IAnnotations(request)

    return _store_in_annotation


def cache(get_key, get_request="request"):
    return volatile.cache(get_key, get_cache=store_in_annotation_of(get_request))


memoize_diy_request = RequestMemo

__all__ = (memoize_diy_request, store_in_annotation_of, cache)



from zope.interface import Interface


class ICacheChooser(Interface):
    def __call__(fun_name):
        """Return a cache with a dict interface based on a dotted
        function name `fun_name`.

        May return None to indicate that there is no cache available.
        """


class IXHTMLCompressor(Interface):
    def compress(string):
        """Expects a valid XHTML Unicode string as input and returns a valid
        XHTML Unicode string.

        Examples of compression are removal of whitespace without a semantical
        meaning, lowercasing of tags and attribute names (which gzip
        compression can take advantage of) or removal of comments.
        """



from pathlib import Path
from setuptools import find_packages
from setuptools import setup


version = "3.0.2"

long_description = (
    f"{Path('README.rst').read_text()}\n"
    f"{(Path('plone') / 'memoize' / 'README.rst').read_text()}\n"
    f"{Path('CHANGES.rst').read_text()}"
)

setup(
    name="plone.memoize",
    version=version,
    description="Decorators for caching the values of functions and methods",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    # Get more strings from
    # https://pypi.org/classifiers/
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Plone",
        "Framework :: Plone :: 6.0",
        "Framework :: Plone :: Core",
        "Framework :: Zope :: 5",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    keywords="plone memoize decorator cache",
    author="Plone Foundation",
    author_email="plone-developers@lists.sourceforge.net",
    url="https://pypi.org/project/plone.memoize",
    license="BSD",
    packages=find_packages(),
    namespace_packages=["plone"],
    include_package_data=True,
    zip_safe=False,
    test_suite="plone.memoize.tests.test_suite",
    python_requires=">=3.8",
    extras_require=dict(
        test=[
            "zope.configuration",
            "zope.publisher",
            "zope.testing",
        ],
    ),
    install_requires=[
        "setuptools",
        "zope.annotation",
        "zope.component",
        "zope.globalrequest",
        "zope.interface",
        "zope.ramcache",
    ],
)



"""Memoize decorator for views.

Stores values in an annotation of the request. See view.rst.
"""
from functools import wraps
from zope.annotation.interfaces import IAnnotations
from zope.globalrequest import getRequest


class ViewMemo:
    key = "plone.memoize"

    def memoize(self, func):
        @wraps(func)
        def memogetter(*args, **kwargs):
            instance = args[0]

            context = getattr(instance, "context", None)
            try:
                request = instance.request
            except AttributeError:
                request = getRequest()

            annotations = IAnnotations(request, {})
            if self.key not in annotations:
                annotations[self.key] = dict()
            cache = annotations[self.key]

            # XXX: Not the most elegant thing in the world; in a Zope 2
            # context, the physical path is a better key, since the id could
            # change if the object is invalidated from the ZODB cache

            try:
                context_id = context.getPhysicalPath()
            except AttributeError:
                context_id = id(context)

            # Note: we don't use args[0] in the cache key, since args[0] ==
            # instance and the whole point is that we can cache different
            # requests

            key = (
                context_id,
                instance.__class__.__name__,
                func.__name__,
                args[1:],
                frozenset(kwargs.items()),
            )
            if key not in cache:
                cache[key] = func(*args, **kwargs)
            return cache[key]

        return memogetter

    def memoize_contextless(self, func):
        def memogetter(*args, **kwargs):
            if args:
                instance = args[0]
            else:
                instance = None

            try:
                request = instance.request
            except AttributeError:
                request = getRequest()

            annotations = IAnnotations(request, {})
            if self.key not in annotations:
                annotations[self.key] = dict()
            cache = annotations[self.key]

            key = (
                instance.__class__.__name__,
                func.__name__,
                args[1:],
                frozenset(kwargs.items()),
            )
            if key not in cache:
                cache[key] = func(*args, **kwargs)
            return cache[key]

        return memogetter


_m = ViewMemo()
memoize = _m.memoize
memoize_contextless = _m.memoize_contextless

__all__ = (memoize, memoize_contextless)



from zope.component.testing import setUp
from zope.component.testing import tearDown
from zope.configuration.xmlconfig import XMLConfig

import doctest
import unittest


optionflags = doctest.REPORT_ONLY_FIRST_FAILURE | doctest.ELLIPSIS


def configurationSetUp(test):
    setUp()
    import zope.component

    XMLConfig("meta.zcml", zope.component)()
    import plone.memoize

    XMLConfig("configure.zcml", plone.memoize)()


def test_suite():
    tests = (
        doctest.DocFileSuite(
            "README.rst",
            package="plone.memoize",
            setUp=configurationSetUp,
            tearDown=tearDown,
            optionflags=optionflags,
        ),
        doctest.DocFileSuite(
            "compress.rst",
            package="plone.memoize",
            setUp=configurationSetUp,
            tearDown=tearDown,
            optionflags=optionflags,
        ),
        doctest.DocFileSuite(
            "forever.rst",
            package="plone.memoize",
            setUp=configurationSetUp,
            tearDown=tearDown,
            optionflags=optionflags,
        ),
        doctest.DocFileSuite(
            "instance.rst",
            package="plone.memoize",
            setUp=configurationSetUp,
            tearDown=tearDown,
            optionflags=optionflags,
        ),
        doctest.DocFileSuite(
            "ram.rst",
            package="plone.memoize",
            setUp=configurationSetUp,
            tearDown=tearDown,
            optionflags=optionflags,
        ),
        doctest.DocFileSuite(
            "request.rst",
            package="plone.memoize",
            setUp=configurationSetUp,
            tearDown=tearDown,
            optionflags=optionflags,
        ),
        doctest.DocFileSuite(
            "view.rst",
            package="plone.memoize",
            setUp=configurationSetUp,
            tearDown=tearDown,
            optionflags=optionflags,
        ),
        doctest.DocFileSuite(
            "volatile.rst",
            package="plone.memoize",
            setUp=configurationSetUp,
            tearDown=tearDown,
            optionflags=optionflags,
        ),
    )

    return unittest.TestSuite(tests)
